package laboratoire4;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

public class MiniMax {
    private static final int MAX_DEPTH = 2;
    private static Random random = new Random();

    public static MiniMaxResult miniMax(PusherBoard board) {
        return miniMax(board, true, 0, 0);
    }

    private static MiniMaxResult miniMax(PusherBoard board, boolean max, int depth, int alphaBeta) {
        int limScore = max ? Integer.MIN_VALUE : Integer.MAX_VALUE;
        List<Pawn> pawns = max ?
                board.getMaxPawns() :
                board.getMinPawns();
//        Collections.shuffle(pawns);

        List<MiniMaxResult> results = new ArrayList<>();

        for (Pawn pawn : pawns) {
            int originalRow = pawn.getRow();
            int originalCol = pawn.getCol();

            for (Pawn.PawnMovement movement : Pawn.PawnMovement.values()) {
                if (pawn.isMoveValid(board, movement)) {
                    pawn.move(movement);

                    int nextAlphaBeta = max ? Math.max(limScore, alphaBeta) : Math.min(limScore, alphaBeta);

                    int score = evaluate(pawn, board, max);
                    if (depth < MAX_DEPTH) {
                        score = miniMax(board, !max, depth + 1, nextAlphaBeta).getScore();
                    }

                    if ((max && score > limScore) ||
                            (!max && score < limScore)) {
                        limScore = score;
                    }

                    MiniMaxResult result = new MiniMaxResult(limScore, pawn, movement);

                    //elagage alphaBeta
//                    if ((max && limScore > alphaBeta) ||
//                            (!max && limScore < alphaBeta)) {
//                        pawn.setRow(originalRow);
//                        pawn.setCol(originalCol);
//
//                        return result;
//                    }

                    results.add(result);
                }

                pawn.setRow(originalRow);
                pawn.setCol(originalCol);
            }
        }

        // Choisir aléatoirement
        int index = random.nextInt(results.size());
        return results.get(index);
    }

    private static int evaluate(Pawn pawn, PusherBoard board, boolean max) {
        int score = didWin(pawn, max) + didCapture(pawn, board, max) + isPushed(pawn);

        int homeRow = pawn.getPlayer() == Player.RED ? 0 : 7;
        if (pawn.getRow() != homeRow) {
            score += 50;
        }

        int direction = max ? 1 : -1;
        return score * direction;
    }

    private static int isPushed(Pawn pawn) {
        if (pawn instanceof Pushed) {
            return 5;
        }

        return 0;
    }

    private static int didWin(Pawn pawn, boolean max) {
        int goal = pawn.getPlayer() == Player.RED ? 7 : 0;

        if (pawn.getRow() == goal) {
            return max ? 1000 : 100000;
        }

        return 0;
    }

    private static int didCapture(Pawn pawn, PusherBoard board, boolean max) {
        Pawn capturedPawn = board.getBoard()[pawn.getRow()][pawn.getCol()];

        if (capturedPawn != null) {
            if (capturedPawn.getPlayer() != pawn.getPlayer()) {
                if (max) {
                    return capturedPawn instanceof Pusher ? 100 : 50;
                } else {
                    return capturedPawn instanceof Pusher ? 10000 : 5000;
                }
            }
        }

        return 0;
    }

    static class MiniMaxResult {
        private final int score;
        private final Pawn pawn;
        private final Pawn.PawnMovement movement;

        public MiniMaxResult(int score, Pawn pawn, Pawn.PawnMovement movement) {
            this.score = score;
            this.pawn = pawn;
            this.movement = movement;
        }

        public int getScore() {
            return score;
        }

        public Pawn getPawn() {
            return pawn;
        }

        public Pawn.PawnMovement getMovement() {
            return movement;
        }

        @Override
        public String toString() {
            return "MiniMaxResult{" +
                    "score=" + score +
                    ", pawn=" + pawn +
                    ", movement=" + movement +
                    '}';
        }
    }
}
